unit uTags;

interface

uses Classes, SysUtils, uTagsMP3, uTagsWMA, uTagsOGG, uTagsWAV,
     Dialogs, DirectShow9, ActiveX;

type
  TTagsData = packed record
    Encoder, Version, Genre, Author, Title, Album, Composer, Url, Track, Year, CopyRight : String[255];
    Comments, Lyrics                                                                     : WideString;
    BitRate, FileSize                                                                    : LongInt;
    Duration                                                                             : Integer;
    VBR                                                                                  : Boolean;
  end;

  TAudioFile = class
    constructor Create;
    destructor Destroy; override;
  private
    // Allgemein
    fSep          : String;
    fPfad         : String;
    fDateiname    : String;
    fOrdner       : String;
    fDateiGroesse : Integer;
    // OGGVorbis-Tags
    fOGGTrack     : Word;
    // aus den ID3-Tags (oder andere Meta-Tags)
    fArtist       : String;
    fTitel        : String;
    fAlbum        : String;
    fTrack        : WideString;
    fYear         : String;
    fGenre        : String;
    fKommentar    : String;
    fLyrics       : String;
    // aus den MPEG-Infos
    fDauer        : Integer;
    fBitrate      : Word;
    fVbr          : Boolean;
    fChannelMode  : String;
    fSamplerate   : Integer;
    fEndung       : String;
    procedure SetNA(filename: String);
    procedure GetMP3Info(filename:String);
    procedure GetWMAInfo(filename:String);
    procedure GetAnotherInfo(filename:String);
    procedure GetOGGInfo(filename:String);
  public
    rTitel  : String;
    rArtist : String;
    Tag     : Boolean;
    // Allgemein
    property Pfad         : String  read fDateiname;
    property Dateiname    : String  read fDateiname;
    property Ordner       : String  read fOrdner;
    property DateiGroesse : Integer read fDateiGroesse;
    property Sep          : String read fSep;
    // OGG-Tags
    property OGGTrack : Word read fOGGTrack;
    // aus den ID3-Tags (oder andere Meta-Tags)
    property Artist    : String     read fArtist;
    property Titel     : String     read fTitel;
    property Album     : String     read fAlbum;
    property Track     : WideString read fTrack;
    property Year      : String     read fYear;
    property Genre     : String     read fGenre;
    property Kommentar : String     read fKommentar;
    property Lyrics    : String     read fLyrics;
    // aus den MPEG-Infos
    property Dauer       : Integer read fDauer;
    property Bitrate     : Word    read fBitrate;
    property vbr         : Boolean read fVbr;
    property ChannelMode : String  read fChannelMode;
    property Samplerate  : Integer read fSamplerate;
    property Endung      : String  read fEndung write fEndung;
    procedure GetAudioData(filename : String);
    procedure SaveMP3Info(filename, titel, artist, album, track, year, comment, genre : String);
    procedure SaveWMAInfo(filename, a, b, c, d, e, f, g : String);
  end;

implementation

{$I-}

constructor TAudioFile.Create;
begin
  inherited Create;
  fDauer       := 0;
  fBitrate     := 0;
  fChannelMode := 'Stereo';
  fSamplerate  := 44100;
  fGenre       := 'Other';
  fYear        := '';
  fEndung      := '';
end;

destructor TAudioFile.Destroy;
begin
  inherited Destroy;
end;

procedure TAudioFile.SetNA(filename : String);
var f : File;
begin
  if FileExists(filename) then
  begin
    try
      AssignFile(f, filename);
      FileMode      := 0;
      Reset(f, 1);
      fDateiGroesse := FileSize(F);
      CloseFile(F);
    except
      fDateiGroesse := 0;
    end;
  end else
    fDateiGroesse := 0;

  fPfad      := filename;
  fOrdner    := ExtractFileDir(filename);
  fDateiname := filename;
  fEndung    := ExtractFileExt(Pfad);

  fTitel       := Dateiname;
  fArtist      := '';
  fAlbum       := '';
  fYear        := '';
  fGenre       := '';
  fDauer       := 0;
  fBitrate     := 0;
  fVbr         := True;
  fChannelMode := 'Unknown';
  fSamplerate  := -1;
  fSep         := '';
end;

procedure TAudioFile.GetMP3Info(filename : String);
var MpegInfo : TMpegInfo;
    ID3v2Tag : TID3V2Tag;
    ID3v1Tag : TID3v1Tag;
    Stream   : TFileStream;
begin
  fPfad      := filename;
  fOrdner    := ExtractFileDir(filename);
  fDateiname := filename;
  if not FileExists(filename) then
    Exit;

  MpegInfo := TMpegInfo.Create;
  ID3v2Tag := TID3V2Tag.Create;
  ID3v1Tag := TID3v1Tag.Create;

  Stream := TFileStream.Create(filename, fmOpenRead or fmShareDenyWrite);
  ID3v1Tag.ReadFromStream(Stream);
  Stream.Seek(0, soBeginning);
  ID3v2Tag.ReadFromStream(Stream);

  if not ID3v2Tag.Exists then
    Stream.Seek(0, sobeginning)
  else
    Stream.Seek(ID3v2Tag.Size, soFromBeginning);

  MpegInfo.LoadFromStream(Stream);
  Stream.free;
  if MpegInfo.FirstHeaderPosition > -1 then
  begin
    if ID3v2Tag.Artist <> '' then
      fArtist := ID3v2Tag.Artist
    else
      if ID3v1Tag.Artist <> '' then
        fArtist := ID3v1Tag.Artist
      else
        fArtist := '';

    if ID3v2Tag.Title <> '' then
      fTitel := ID3v2Tag.Title
    else
      if ID3v1Tag.Title <> '' then
        fTitel := ID3v1Tag.Title
      else
        fTitel := Dateiname;

    if ID3v2Tag.Album <> '' then
      fAlbum := ID3v2Tag.Album
    else
      if ID3v1Tag.Album <> '' then
        fAlbum := ID3v1Tag.Album
      else
        fAlbum := '';

    if ID3v2Tag.Track <> '' then
      fTrack := ID3v2Tag.Track
    else
      if ID3v1Tag.Track <> '' then
        fTrack := ID3v1Tag.Track
      else
        fTrack := 'N/A';

    if ID3v2Tag.Year <> '' then
      fYear := ID3v2Tag.Year
    else
      if ID3v1Tag.Year <> '' then
        fYear := ID3v1Tag.Year
      else
        fYear := '';

    if ID3v2Tag.genre <> '' then
      fGenre := ID3v2Tag.Genre
    else
      fGenre := ID3v1Tag.Genre;

    if ID3v2Tag.Comment <> '' then
      fKommentar := ID3v2Tag.Comment
    else
      if ID3v1Tag.Comment <> '' then
        fKommentar := ID3v1Tag.Comment
      else
        fKommentar := '';
    // Liedtexte knnen nur im v2-Tag gespeichert werden.
    fLyrics := ID3v2Tag.GetLyrics('*', '*');
    // aus den MPEG-Infos
    fDauer        := MpegInfo.dauer;
    fBitrate      := MpegInfo.bitrate;
    fVbr          := MpegInfo.vbr;
    fDateiGroesse := MpegInfo.FileSize;
    fSamplerate   := MpegInfo.samplerate;
    case MpegInfo.channelmode of
      0 : fChannelMode := 'Stereo';
      1 : fChannelMode := 'Joint Stereo';
      2 : fChannelMode := 'Dual Channel Stereo';
      3 : fChannelMode := 'Mono';
      else
        fChannelMode := 'Unknown';
    end;
  end else
    SetNA(filename);

  MpegInfo.Free;
  Id3v2Tag.Free;
  Id3v1Tag.Free;
  fPfad := Filename;
  fSep  := ' - ';
end;

procedure TAudioFile.SAVEMP3Info(filename, titel, artist, album, track, year, comment, genre : String);
var MpegInfo : TMpegInfo;
    ID3v2Tag : TID3V2Tag;
    ID3v1Tag : TID3v1Tag;
    Stream   : TFileStream;
begin
  if not FileExists(filename) then
    Exit;

  MpegInfo := TMpegInfo.Create;
  ID3v2Tag := TID3V2Tag.Create;
  ID3v1Tag := TID3v1Tag.Create;

  // Datei als Stream ffnen und Daten auslesen
  Stream := TFileStream.Create(filename, fmOpenRead or fmShareDenyWrite);
  ID3v1Tag.ReadFromStream(Stream);
  Stream.Seek(0, soBeginning);
  ID3v2Tag.ReadFromStream(Stream);

  if not ID3v2Tag.Exists then
    Stream.Seek(0, soBeginning)
  else
    Stream.Seek(ID3v2Tag.Size, soFromBeginning);

  MpegInfo.LoadFromStream(Stream);
  Stream.Free;

  if MpegInfo.FirstHeaderPosition > -1 then
  begin
    if ID3v2Tag.Artist <> '' then
      ID3v2Tag.Artist := Artist
    else
      if ID3v1Tag.Artist <> '' then
        ID3v1Tag.Artist := Artist
      else
        ID3v1Tag.Artist := '';

    if ID3v2Tag.Title <> '' then
      ID3v2Tag.Title := Titel
    else
      if ID3v1Tag.Title <> '' then
        ID3v1Tag.Title := Titel;

    if ID3v2Tag.Album <> '' then
      ID3v2Tag.Album := Album
    else
      if ID3v1Tag.Album <> '' then
        ID3v1Tag.Album := Album
      else
        ID3v1Tag.Album := '';

    if ID3v2Tag.Track <> '' then
      ID3v2Tag.Track := Track
    else
      if ID3v1Tag.Track <> '' then
        ID3v1Tag.Track := Track
      else
        ID3v1Tag.Track := 'N/A';

    if ID3v2Tag.Year <> '' then
      ID3v2Tag.Year := Year
    else
      if ID3v1Tag.Year <> '' then
        ID3v1Tag.Year := Year
      else
        ID3v1Tag.Year := '';

    if ID3v2Tag.Genre <> '' then
      ID3v2Tag.Genre := Genre
    else
      ID3v1Tag.Genre := Genre;

    if ID3v2Tag.Comment <> '' then
      ID3v2Tag.Comment := Comment
    else
      if ID3v1Tag.Comment <> '' then
        ID3v1Tag.Comment := Comment
      else
        ID3v1Tag.Comment := '';
    ID3v1Tag.WriteToFile(filename);
    ID3v2Tag.WriteToFile(filename)
  end;
  MpegInfo.Free;
  Id3v2Tag.Free;
  Id3v1Tag.Free;
end;

procedure TAudioFile.GetWMAInfo(filename : String);
var TagsData : PTagsData;
begin
  New(TagsData);
  try
    OpenWMAFile(filename, TagsData);
    fArtist    := TagsData.Author;
    fTitel     := TagsData.Title;
    fAlbum     := TagsData.Album;
    fGenre     := TagsData.Genre;
    fYear      := TagsData.Year;
    fTrack     := TagsData.Track;
    fKommentar := TagsData.Comments;
    fDauer     := TagsData.Duration;
  finally
    DisPose(TagsData);
  end;
  fPfad      := filename;
  fOrdner    := ExtractFileDir(filename);
  fDateiname := filename;
  rTitel     := fTitel;
  rArtist    := fArtist;
  fSep       := ' - ';
end;

procedure TAudioFile.SaveWMAInfo(filename, a, b, c, d, e, f, g : String);
var TagsData : PTagsData;
begin
  New(TagsData);
  try
    TagsData.Title    := a;
    TagsData.Author   := b;
    TagsData.Album    := c;
    TagsData.Genre    := d;
    TagsData.Year     := e;
    TagsData.Track    := f;
    TagsData.Comments := g;
    SaveWMAFile(TagsData);
  finally
    Dispose(TagsData);
  end;
end;

procedure TAudioFile.GetOGGInfo(filename : String);
var OGG : TOGGVorbis;
begin
  OGG := TOGGVorbis.Create;
  OGG.ReadFromFile(filename);
  fArtist    := OGG.Artist;
  fTitel     := OGG.Title;
  fAlbum     := OGG.Album;
  fOGGTrack  := OGG.Track;
  fYear      := OGG.Date;
  fGenre     := OGG.Genre;
  fKommentar := OGG.Comment;
  fDauer     := Round(OGG.Duration);
  OGG.Free;
  fPfad      := filename;
  fOrdner    := ExtractFileDir(filename);
  fDateiname := filename;
  fSep       := ' - ';
end;

procedure TAudioFile.GetAnotherInfo(filename : String);
  function GetMediaLength(const aFileName : WideString) : Double;
  var MediaDet : IMediaDet;
      hr       : HResult;
  begin
    hr := CoCreateInstance(CLSID_MediaDet, nil, CLSCTX_INPROC_SERVER, IID_IMediaDet, MediaDet);
    if Succeeded(hr) then
    begin
      MediaDet.put_Filename(aFilename);
      MediaDet.get_StreamLength(Result);
    end else
      Result := -1;
    MediaDet := nil;
  end;
var f : String;
    s : String;
    p : Integer;
begin
  if Trim(filename) = '' then
    Exit;
  f := Copy(ExtractFileName(filename), 0, Length(ExtractFileName(filename))-Length(ExtractFileExt(filename)));
  s := FloatToStr(GetMediaLength(filename));
  p := Pos(',', s);
  if not p <= -1 then
    s := Copy(s, 0, p-1);
  fDauer     := StrToInt(s);
  fArtist    := f;
  fTitel     := '';
  fPfad      := filename;
  fOrdner    := ExtractFileDir(filename);
  fDateiname := filename;
  fSep       := '';
end;

procedure TAudioFile.GetAudioData(filename : String);
var s : String;
begin    
  s := AnsiLowerCase(ExtractFileExt(filename));
  try
    if (s = '.mp3') or (s = '.mp2') or (s = '.mp1') then
      GetMP3Info(Filename)
    else
    begin
      if s = '.wma' then
        GetWMAInfo(filename)
      else if s = '.wav' then
        GetAnotherInfo(filename)
      else if s = '.OGG' then
        GetOGGInfo(filename)
      else if s = '.aiff' then
        GetAnotherInfo(filename)
      else if s = '.aif' then
        GetAnotherInfo(filename)
      else if s = '.mpg' then
        GetAnotherInfo(filename)
      else if s = '.mpeg' then
        GetAnotherInfo(filename)
      else if s = '.avi' then
        GetAnotherInfo(filename)
      else if s = '.dat' then
        GetAnotherInfo(filename)
      else if s = '.wmv' then
        GetAnotherInfo(filename);
    end;
  except
    SetNA(filename);
  end;
end;

end.