unit uTagsWMA;

interface

uses
  SysUtils, Windows, Dialogs, uTypes;

type
  TTagsData = packed record
    Encoder, Version, Genre, Author, Title, Album, Composer, Url, Track, Year, CopyRight : String[255];
    Comments, Lyrics                                                                     : WideString;
    BitRate, FileSize                                                                    : LongInt;
    Duration                                                                             : Integer;
    VBR                                                                                  : Boolean;
  end;
  PTagsData = ^TTagsData;

procedure OpenWMAFile(FileName : String; TagsData : PTagsData);
procedure SaveWMAFile(TagsData : PTagsData);
function  DurationToStr(Duration : Integer; ShowMs : Boolean): String;

implementation

const
  IID_IWMMetadataEditor : TGUID = '{96406bd9-2b2b-11d3-b36b-00c04f6108ff}';
  IID_IWMHeaderInfo     : TGUID = '{96406bda-2b2b-11d3-b36b-00c04f6108ff}';
  IID_IWMHeaderInfo3    : TGUID = '{15CC68E3-27CC-4ecd-B222-3F5D02D80BD5}';
  IID_IWMHeaderInfo2    : TGUID = '{15cf9781-454e-482e-b393-85fae487a810}';

  LibName = 'WMVCORE.DLL';

var
  fFileName : PWideChar;

type
  WMT_CODEC_INFO_TYPE = LongWord;

  WMT_ATTR_DATATYPE = (WMT_TYPE_DWORD, WMT_TYPE_String, WMT_TYPE_BINARY, WMT_TYPE_BOOL, WMT_TYPE_QWORD, WMT_TYPE_WORD, WMT_TYPE_GUID);

  TWMTAttrDataType = WMT_ATTR_DATATYPE;

  IWMMetadataEditor = interface(IUnknown)
    ['{96406BD9-2B2B-11d3-B36B-00C04F6108FF}']
    function Open(pwszFilename : PWideChar) : HRESULT; stdcall;
    function Close : HRESULT; stdcall;
    function Flush : HRESULT; stdcall;
  end;

  IWMHeaderInfo = interface(IUnknown)
    ['{96406BDA-2B2B-11d3-B36B-00C04F6108FF}']
    function GetAttributeCount(wStreamNum : Word; out pcAttributes : Word): HRESULT; stdcall;
    function GetAttributeByIndex(wIndex : Word; var pwStreamNum : Word; pwszName : PWideChar; var pcchNameLen : Word;
      out pType : TWMTAttrDataType; pValue : PBYTE; var pcbLength : Word) : HRESULT; stdcall;
    function GetAttributeByName(var pwStreamNum : Word; pszName : PWideChar; out pType : TWMTAttrDataType;
      pValue : PBYTE; var pcbLength : Word) : HRESULT; stdcall;
    function SetAttribute(wStreamNum : Word; pszName : PWideChar; Type_ : TWMTAttrDataType; {in} pValue : PBYTE;
      cbLength : Word) : HRESULT; stdcall;
    function GetMarkerCount(out pcMarkers : Word) : HRESULT; stdcall;
    function GetMarker(wIndex : Word; pwszMarkerName : PWideChar; var pcchMarkerNameLen : Word;
      out pcnsMarkerTime : Int64) : HRESULT; stdcall;
    function AddMarker(pwszMarkerName : PWideChar; cnsMarkerTime : Int64) : HRESULT; stdcall;
    function RemoveMarker(wIndex : Word) : HRESULT; stdcall;
    function GetScriptCount(out pcScripts : Word) : HRESULT; stdcall;
    function GetScript(wIndex : Word; pwszType : PWideChar; var pcchTypeLen : Word; pwszCommand : PWideChar;
      var pcchCommandLen : Word; out pcnsScriptTime : Int64) : HRESULT; stdcall;
    function AddScript(pwszType, pwszCommand : PWideChar; cnsScriptTime : Int64) : HRESULT; stdcall;
    function RemoveScript(wIndex : Word) : HRESULT; stdcall;
  end;

  IWMHeaderInfo2 = interface(IWMHeaderInfo)
    ['{15CF9781-454E-482e-B393-85FAE487A810}']
    function GetCodecInfoCount(out pcCodecInfos : LongWord) : HRESULT; stdcall;
    function GetCodecInfo(wIndex : LongWord; var pcchName : Word; pwszName : PWideChar; var pcchDescription : Word;
      pwszDescription : PWideChar; out pCodecType : WMT_CODEC_INFO_TYPE; var pcbCodecInfo : Word;
      pbCodecInfo : PBYTE) : HRESULT; stdcall;
  end;

  IWMHeaderInfo3 = interface(IWMHeaderInfo2)
    ['{15CC68E3-27CC-4ecd-B222-3F5D02D80BD5}']
    function GetAttributeCountEx(wStreamNum : Word; out pcAttributes : Word) : HRESULT; stdcall;
    function GetAttributeIndices(wStreamNum : Word; pwszName : PWideChar; pwLangIndex : PWORD; pwIndices : PWORD;
      var pwCount : Word) : HRESULT; stdcall;
    function GetAttributeByIndexEx(wStreamNum : Word; wIndex : Word; pwszName : PWideChar; var pwNameLen : Word;
      out pType : TWMTAttrDataType; out pwLangIndex : Word; pValue : PBYTE; var pdwDataLength : LongWord) : HRESULT;
      stdcall;
    function ModifyAttribute(wStreamNum : Word; wIndex : Word; Type_ : TWMTAttrDataType; wLangIndex : Word;
      pValue : PBYTE; dwLength : LongWord) : HRESULT; stdcall;
    function AddAttribute(wStreamNum : Word; pszName : PWideChar; out pwIndex : Word; Type_ : TWMTAttrDataType;
      wLangIndex : Word; pValue : PBYTE; dwLength : LongWord) : HRESULT; stdcall;
    function DeleteAttribute(wStreamNum, wIndex : Word) : HRESULT; stdcall;
    function AddCodecInfo(pwszName : PWideChar; pwszDescription : PWideChar; codecType : WMT_CODEC_INFO_TYPE;
      cbCodecInfo : Word; pbCodecInfo : PBYTE) : HRESULT; stdcall;
  end;

var _WMCreateEditor : function(out ppEditor : IWMMetadataEditor) : HRESULT; stdcall;
    DllHandle       : THandle;
    ppEditor        : IWMMetadataEditor;
    ppHeaderInfo3   : IWMHeaderInfo3;
    spSource        : TDynStrArray;

function Split(const Source, Delimiter: String): TDynStrArray;
var spCount, spPos, spLength : Integer;
    sTemp                    : String;
    aSplit                   : TDynStrArray;
begin
  sTemp    := Source;
  spCount  := 0;
  spLength := Length(Delimiter)-1;
  repeat
    spPos := Pos(Delimiter, sTemp);
    if spPos = 0 then
      Break
    else
    begin
      inc(spCount);
      SetLength(aSplit, spCount);
      aSplit[spCount-1] := Copy(sTemp, 1, spPos-1);
      Delete(sTemp, 1, spPos+spLength);
    end;
  until
    False;
  if Length(sTemp) > 0 then
  begin
    inc(spCount);
    SetLength(aSplit, spCount);
    aSplit[spCount-1] := sTemp;
  end;
  Result := aSplit;
end;

procedure ResetTagsData(TagsData : PTagsData);
begin
  TagsData.Duration  := 0;
  TagsData.BitRate   := 0;
  TagsData.FileSize  := 0;
  TagsData.Encoder   := '';
  TagsData.Version   := '';
  TagsData.Genre     := '';
  TagsData.Author    := '';
  TagsData.Composer  := '';
  TagsData.Title     := '';
  TagsData.Album     := '';
  TagsData.Year      := '';
  TagsData.Track     := '';
  TagsData.Url       := '';
  TagsData.Comments  := '';
  TagsData.Lyrics    := '';
  TagsData.CopyRight := '';
end;

function GetAttribIndex(Attrib : String) : Integer;
var AttribName : PWideChar;
    P, Lang    : PWord;
    AttribLen  : Word;
begin
  Result := -1;

  AttribName := StringToOleStr(Attrib);
  Lang := PWord(0);

  ppHeaderInfo3.GetAttributeIndices($FFFF, AttribName, Lang, nil, AttribLen);

  if AttribLen <> 0 then
  begin
    P := AllocMem(AttribLen);
    try
      ppHeaderInfo3.GetAttributeIndices($FFFF, AttribName, Lang, P, AttribLen);
      Result := PWord(P)^;
    finally
      FreeMem(P);
    end;
  end;
end;

function ShowTags(var FileName : PWideChar; var TagsData : PTagsData) : Boolean; stdcall;
var AttribStream, AttributeCount, AttribLen : Word;
    AttribName                              : PWideChar;
    pType                                   : TWMTAttrDataType;
    pValue                                  : PByte;
    DataLen                                 : Word;
    i, nSize, nVal                          : Integer;
    HR                                      : HRESULT;
    P                                       : PInteger;
begin
  DllHandle := LoadLibrary(LibName);
  if DllHandle = 0 then
  begin
    ShowMessage(Format('Could not Load Library ($%x).', [DllHandle]));
    Result := False;
    Exit;
  end else
    _WMCreateEditor := GetProcAddress(DllHandle, 'WMCreateEditor');

  ResetTagsData(TagsData);
  if not FileExists(FileName) then
  begin
    raise Exception.Create(Format('The file' + #13#10 + '%S' + #13#10 + 'does''nt exists!', [FileName]));
    Exit;
  end;

  try
    HR := _WMCreateEditor(ppEditor);
    if Failed(HR) then
    begin
      Result := False;
      raise Exception.Create(Format('Could not create Metadata Editor ($%x).', [HR]));
      Exit;
    end;
    ppEditor.Open(FileName);
    HR := ppEditor.QueryInterface(IID_IWMHeaderInfo3, ppHeaderInfo3);
    if Failed(HR) then
    begin
      Result := False;
      raise Exception.Create(Format('Could not QI for IWMHeaderInfo3 ($%x).', [HR]));
      Exit;
    end;
    Result       := True;
    AttribStream := 0;
    ppHeaderInfo3.GetAttributeCount(AttribStream, AttributeCount);
    for i := 0 to AttributeCount - 1 do
    begin
      ppHeaderInfo3.GetAttributeByIndex(Word(i), AttribStream, nil, AttribLen, pType, nil, DataLen);

      nSize      := 255;
      AttribName := AllocMem(nSize*2);
      AttribLen  := nSize;
      PValue     := AllocMem(DataLen);

      ppHeaderInfo3.GetAttributeByIndex(Word(i), AttribStream, PWideChar(AttribName), AttribLen, pType, PByte(PValue), DataLen);
      if pType = WMT_TYPE_String then
      begin
        if UpperCase(AttribName) = 'WM/ENCODEDBY' then
          TagsData.Encoder := WideString(PValue);
          spSource := Split(TagsData.Encoder, #0);
          if spSource <> nil then
            TagsData.Encoder := spSource[0];

        if UpperCase(AttribName) = 'WMFSDKVERSION' then
          TagsData.Version := WideString(PValue);
        spSource := Split(TagsData.Version, #0);
        if spSource <> nil then
          TagsData.Version := spSource[0];

        if UpperCase(AttribName) = 'WM/GENRE' then
          TagsData.Genre := WideString(PValue);
        spSource := Split(TagsData.Genre, #0);
        if spSource <> nil then
          TagsData.Genre := spSource[0];

        if UpperCase(AttribName) = 'AUTHOR' then
          TagsData.Author := WideString(PValue);
        spSource := Split(TagsData.Author, #0);
        if spSource <> nil then
          TagsData.Author := spSource[0];

        if UpperCase(AttribName) = 'TITLE' then
          TagsData.Title := WideString(PValue);
        spSource := Split(TagsData.Title, #0);
        if spSource <> nil then
          TagsData.Title := spSource[0];

        if UpperCase(AttribName) = 'WM/ALBUMTITLE' then
          TagsData.Album := WideString(PValue);
        spSource := Split(TagsData.Album, #0);
        if spSource <> nil then
          TagsData.Album := spSource[0];

        if UpperCase(AttribName) = 'WM/COMPOSER' then
          TagsData.Composer := WideString(PValue);
        spSource := Split(TagsData.Composer, #0);
        if spSource <> nil then
          TagsData.Composer := spSource[0];

        if UpperCase(AttribName) = 'WM/URL' then
          TagsData.Url := WideString(PValue);
          spSource := Split(TagsData.Url, #0);
        if spSource <> nil then
          TagsData.Url := spSource[0];

        if UpperCase(AttribName) = 'WM/YEAR' then
          TagsData.Year := WideString(PValue);
          spSource := Split(TagsData.Year, #0);
        if spSource <> nil then
          TagsData.Year := spSource[0];

        if UpperCase(AttribName) = 'WM/TRACKNUMBER' then
          TagsData.Track := WideString(PValue);
          spSource := Split(TagsData.Track, #0);
        if spSource <> nil then
          TagsData.Track := spSource[0];

        if UpperCase(AttribName) = 'DESCRIPTION' then
          TagsData.Comments := WideString(PValue);

        if UpperCase(AttribName) = 'WM/LYRICS' then
          TagsData.Lyrics := PWideChar(PValue);

        if UpperCase(AttribName) = 'COPYRIGHT' then
          TagsData.CopyRight := WideString(PValue);
        spSource := Split(TagsData.CopyRight, #0);
        if spSource <> nil then
          TagsData.CopyRight := spSource[0];
      end;

      if pType = WMT_TYPE_BOOL then
        if UpperCase(AttribName) = 'ISVBR' then
        begin
          nVal         := pValue^;
          TagsData.VBR := (nVal <> 0);
        end;

      if pType = WMT_TYPE_DWORD then
      begin
        P    := PInteger(PValue);
        nVal := P^;

        if UpperCase(AttribName) = 'BITRATE' then
          TagsData.BitRate := nVal;
      end;

      if pType = WMT_TYPE_QWORD then
      begin
        P    := PInteger(PValue);
        nVal := P^;

        if UpperCase(AttribName) = 'FILESIZE' then
          TagsData.FileSize := nVal;
      end;
      // it seams that sometimes the Tag Header will Show a wrong Duration
      // This is a simple Calculation
      if (TagsData.FileSize > 0) and (TagsData.BitRate > 0) then
        TagsData.Duration := Round((TagsData.FileSize * 8)/(TagsData.BitRate / 1000 + 0.5)/1000)
      else
        TagsData.Duration := 0;

      FreeMem(AttribName);
      FreeMem(PValue);
    end;
    ppHeaderInfo3 := nil;
  finally
    ppEditor.Close;
    ppEditor := nil;
    if DllHandle <> 0 then
    begin
      FreeLibrary(DllHandle);
      _WMCreateEditor := nil;
    end;
  end;
end;

function SaveTags(var FileName : PWideChar; var TagsData : PTagsData) : Boolean; stdcall;
var nIndex, nLength    : Integer;
    AttribName, pValue : PWideChar;
    pwIndex            : Word;
    HR                 : HRESULT;
begin
  if not FileExists(FileName) then
  begin
    raise Exception.Create(Format('The file' + #13#10 + '%S' + #13#10 + 'does''nt exists!', [FileName]));
    Exit;
  end;

  DllHandle := LoadLibrary(LibName);
  if DllHandle = 0 then
  begin
    ShowMessage(Format('Could not Load Library ($%x).', [DllHandle]));
    Result := False;
    Exit;
  end else
    _WMCreateEditor := GetProcAddress(DllHandle, 'WMCreateEditor');

  Result := True;

  HR := _WMCreateEditor(ppEditor);
  if Failed(HR) then
  begin
    raise Exception.Create(Format('Could not create Metadata Editor ($%x).', [HR]));
    Exit;
  end;

  HR := ppEditor.QueryInterface(IID_IWMHeaderInfo3, ppHeaderInfo3);
  if Failed(HR) then
  begin
    raise Exception.Create(Format('Could not QI for IWMHeaderInfo3 ($%x).', [HR]));
    Exit;
  end;

  ppEditor.Open(FileName);
  try
    AttribName := 'WM/EncodedBy';
    nIndex     := GetAttribIndex(AttribName);
    pValue     := StringToOleStr(TagsData.Encoder);
    nLength    := Length(pValue)*2;
    if nIndex >= 0 then
    begin
      if TagsData.Encoder <> '' then
        ppHeaderInfo3.ModifyAttribute(0, nIndex, WMT_TYPE_String, 0, PByte(pValue), nLength)
      else
        ppHeaderInfo3.DeleteAttribute(0, nIndex);
    end else
    if nLength <> 0 then
      ppHeaderInfo3.AddAttribute(0, AttribName, pwIndex, WMT_TYPE_String, 0, PByte(pValue), nLength);

    AttribName := 'WM/Genre';
    nIndex     := GetAttribIndex(AttribName);
    pValue     := StringToOleStr(TagsData.Genre);
    nLength    := Length(pValue)*2;
    if nIndex >= 0 then
    begin
      if TagsData.Genre <> '' then
        ppHeaderInfo3.ModifyAttribute(0, nIndex, WMT_TYPE_String, 0, PByte(pValue), nLength)
      else
        ppHeaderInfo3.DeleteAttribute(0, nIndex);
    end else
    if nLength <> 0 then
      ppHeaderInfo3.AddAttribute(0, AttribName, pwIndex, WMT_TYPE_String, 0, PByte(pValue), nLength);


    AttribName := 'Author';
    nIndex := GetAttribIndex(AttribName);
    pValue := StringToOleStr(TagsData.Author);
    nLength := Length(pValue) * 2;
    if nIndex >= 0 then
    begin
      if TagsData.Author <> '' then
        ppHeaderInfo3.ModifyAttribute(0, nIndex, WMT_TYPE_String, 0, PByte(pValue), nLength)
      else
        ppHeaderInfo3.DeleteAttribute(0, nIndex);
    end else
    if nLength <> 0 then
      ppHeaderInfo3.AddAttribute(0, AttribName, pwIndex, WMT_TYPE_String, 0, PByte(pValue), nLength);


    AttribName := 'WM/Composer';
    nIndex     := GetAttribIndex(AttribName);
    pValue     := StringToOleStr(TagsData.Composer);
    nLength    := Length(pValue)*2;
    if nIndex >= 0 then
    begin
      if TagsData.Composer <> '' then
        ppHeaderInfo3.ModifyAttribute(0, nIndex, WMT_TYPE_String, 0, PByte(pValue), nLength)
      else
        ppHeaderInfo3.DeleteAttribute(0, nIndex);
    end else
    if nLength <> 0 then
      ppHeaderInfo3.AddAttribute(0, AttribName, pwIndex, WMT_TYPE_String, 0, PByte(pValue), nLength);

    AttribName := 'Title';
    nIndex     := GetAttribIndex(AttribName);
    pValue     := StringToOleStr(TagsData.Title);
    nLength    := Length(pValue)*2;
    if nIndex >= 0 then
    begin
      if TagsData.Title <> '' then
        ppHeaderInfo3.ModifyAttribute(0, nIndex, WMT_TYPE_String, 0, PByte(pValue), nLength)
      else
        ppHeaderInfo3.DeleteAttribute(0, nIndex);
    end else
    if nLength <> 0 then
      ppHeaderInfo3.AddAttribute(0, AttribName, pwIndex, WMT_TYPE_String, 0, PByte(pValue), nLength);

    AttribName := 'WM/AlbumTitle';
    nIndex     := GetAttribIndex(AttribName);
    pValue     := StringToOleStr(TagsData.Album);
    nLength    := Length(pValue)*2;
    if nIndex >= 0 then
    begin
      if TagsData.Album <> '' then
        ppHeaderInfo3.ModifyAttribute(0, nIndex, WMT_TYPE_String, 0, PByte(pValue), nLength)
      else
        ppHeaderInfo3.DeleteAttribute(0, nIndex);
    end else
    if nLength <> 0 then
      ppHeaderInfo3.AddAttribute(0, AttribName, pwIndex, WMT_TYPE_String, 0, PByte(pValue), nLength);

    AttribName := 'WM/Year';
    nIndex     := GetAttribIndex(AttribName);
    pValue     := StringToOleStr(TagsData.Year);
    nLength    := Length(pValue)*2;
    if nIndex >= 0 then
    begin
      if TagsData.Year <> '' then
        ppHeaderInfo3.ModifyAttribute(0, nIndex, WMT_TYPE_String, 0, PByte(pValue), nLength)
      else
        ppHeaderInfo3.DeleteAttribute(0, nIndex);
    end else
    if nLength <> 0 then
      ppHeaderInfo3.AddAttribute(0, AttribName, pwIndex, WMT_TYPE_String, 0, PByte(pValue), nLength);

    AttribName := 'WM/TrackNumber';
    nIndex     := GetAttribIndex(AttribName);
    pValue     := StringToOleStr(TagsData.Track);
    nLength    := Length(pValue)*2;
    if nIndex >= 0 then
    begin
      if TagsData.Track <> '' then
        ppHeaderInfo3.ModifyAttribute(0, nIndex, WMT_TYPE_String, 0, PByte(pValue), nLength)
      else
        ppHeaderInfo3.DeleteAttribute(0, nIndex);
    end else
    if nLength <> 0 then
      ppHeaderInfo3.AddAttribute(0, AttribName, pwIndex, WMT_TYPE_String, 0, PByte(pValue), nLength);

    AttribName := 'WM/URL';
    nIndex := GetAttribIndex(AttribName);
    pValue := StringToOleStr(TagsData.Url);
    nLength := Length(pValue)*2;
    if nIndex >= 0 then
    begin
      if TagsData.Url <> '' then
        ppHeaderInfo3.ModifyAttribute(0, nIndex, WMT_TYPE_String, 0, PByte(pValue), nLength)
      else
        ppHeaderInfo3.DeleteAttribute(0, nIndex);
    end else
    if nLength <> 0 then
      ppHeaderInfo3.AddAttribute(0, AttribName, pwIndex, WMT_TYPE_String, 0, PByte(pValue), nLength);


    AttribName := 'Copyright';
    nIndex     := GetAttribIndex(AttribName);
    pValue     := StringToOleStr(TagsData.Copyright);
    nLength    := Length(pValue)*2;
    if nIndex >= 0 then
    begin
      if TagsData.Copyright <> '' then
        ppHeaderInfo3.ModifyAttribute(0, nIndex, WMT_TYPE_String, 0, PByte(pValue), nLength)
      else
        ppHeaderInfo3.DeleteAttribute(0, nIndex);
    end else
    if nLength <> 0 then
      ppHeaderInfo3.AddAttribute(0, AttribName, pwIndex, WMT_TYPE_String, 0, PByte(pValue), nLength);


    AttribName := 'Description';
    nIndex     := GetAttribIndex(AttribName);
    pValue     := StringToOleStr(TagsData.Comments);
    nLength    := Length(pValue)*2;
    if nIndex >= 0 then
    begin
      if TagsData.Comments <> '' then
        ppHeaderInfo3.ModifyAttribute(0, nIndex, WMT_TYPE_String, 0, PByte(pValue), nLength)
      else
        ppHeaderInfo3.DeleteAttribute(0, nIndex);
    end else
    if nLength <> 0 then
      ppHeaderInfo3.AddAttribute(0, AttribName, pwIndex, WMT_TYPE_String, 0, PByte(pValue), nLength);

    AttribName := 'WM/Lyrics';
    nIndex     := GetAttribIndex(AttribName);
    pValue     := StringToOleStr(TagsData.Lyrics);
    nLength    := Length(pValue)*2;
    if nIndex >= 0 then
    begin
      if TagsData.Lyrics <> '' then
        ppHeaderInfo3.ModifyAttribute(0, nIndex, WMT_TYPE_String, 0, PByte(pValue), nLength)
      else
        ppHeaderInfo3.DeleteAttribute(0, nIndex);
    end else
    if nLength <> 0 then
      ppHeaderInfo3.AddAttribute(0, AttribName, pwIndex, WMT_TYPE_String, 0, PByte(pValue), nLength);
    ppEditor.Flush;
  finally
    ppHeaderInfo3 := nil;

    ppEditor.Close;
    ppEditor := nil;

    if DllHandle <> 0 then
    begin
      FreeLibrary(DllHandle);
      _WMCreateEditor := nil;
    end;
  end;
end;

procedure OpenWMAFile(FileName : String; TagsData : PTagsData);
begin
  if not FileExists(FileName) then
    Exit;

  fFileName := StringToOleStr(FileName);

  if not ShowTags(fFileName, TagsData) Then
    ShowMessage('Could Not Load Tags!');
end;

procedure SaveWMAFile(TagsData : PTagsData);
begin
  if not SaveTags(fFileName, TagsData) then
    ShowMessage('Could not save Tags!');
end;

function DurationToStr(Duration : Integer; ShowMs : Boolean) : String;
begin
  if ShowMs then
    Result := Format('%d:%2.2d,%3.3d', [(Abs(Duration div 60)), Abs(Duration mod 60), Abs(Duration mod 1000)])
  else
    Result := Format('%d:%2.2d', [(Abs(Duration div 60)), Abs(Duration mod 60)]);
end;

end.

